library pam_flutter;

import 'dart:convert';
import 'dart:core';
import 'package:device_info/device_info.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:http/http.dart';
import 'package:package_info/package_info.dart';
import 'package:pam_flutter/models/pam_message_model.dart';
import 'package:pam_flutter/pages/consent_page.dart';
import 'package:pam_flutter/pages/notification_page.dart';
import 'package:pam_flutter/pam_notification.dart';
import 'package:pam_flutter/pam_queue_manager.dart';
import 'package:pam_flutter/types.dart';
import 'package:pam_flutter/utils/requester.dart';
import 'package:shared_preferences/shared_preferences.dart';
import 'package:flutter/services.dart' show rootBundle;
import 'dart:io' show Platform;
import 'package:uuid/uuid.dart';

class Pam {
  Pam._();

  static late BuildContext _context;
  static bool _isAppReady = false;
  static late Map<String, dynamic> _config;
  static late SharedPreferences _sharedPreferences;
  static String? _appPrefix;
  static String? _pamServer;
  static String? _publicDBAlias;
  static String? _loginDBAlias;
  static String? _baseAPI;
  static String? _trackingConsentId;
  static String? _contactingConsentId;
  static late bool _enableLog;
  static late PamQueueManager _pamQueueManager;
  static List<void Function(Map<String, dynamic>)> onConsentListener =
      List<void Function(Map<String, dynamic>)>.empty(growable: true);
  static List<void Function(Map<String, dynamic>)> onTokenListener =
      List<void Function(Map<String, dynamic>)>.empty(growable: true);
  static List<void Function(Map<String, dynamic>)> onMessageOpenedListener =
      List<void Function(Map<String, dynamic>)>.empty(growable: true);
  static List<void Function(Map<String, dynamic>)> onMessageListener =
      List<void Function(Map<String, dynamic>)>.empty(growable: true);
  static List<Map<String, dynamic>> _pendingMessage =
      List<Map<String, dynamic>>.empty(growable: true);
  static late String _platformVersion;
  static late String _osVersion;
  static late String _appVersion;
  static String? _contactId;
  static String? _customerId;
  static Widget Function(String id)? _productDetail;

  static Future<void> initial({bool enableLog = false}) async {
    _isAppReady = false;
    _enableLog = enableLog;
    _pamQueueManager = PamQueueManager(callback: _postTracker);

    await _initialConfig();
    await _initialSharedPreferences();
    await _initialAppInfo();
    await PamNotification.pamInitial();
  }

  static Future<void> _initialConfig() async {
    try {
      String raw = await rootBundle.loadString("pam-config.json");
      _config = json.decode(raw);
    } catch (e) {
      throw (e);
    }
  }

  static Future<void> _initialSharedPreferences() async {
    _sharedPreferences = await SharedPreferences.getInstance();
  }

  static Future<void> _initialAppInfo() async {
    DeviceInfoPlugin deviceInfo = DeviceInfoPlugin();
    PackageInfo packageInfo = await PackageInfo.fromPlatform();
    _appVersion = "${packageInfo.packageName} (${packageInfo.version})";

    if (Platform.isAndroid) {
      AndroidDeviceInfo androidInfo = await deviceInfo.androidInfo;
      _osVersion = "Android: ${androidInfo.version.sdkInt}";
      _platformVersion =
          "Android: ${androidInfo.version.release} (SDK ${androidInfo.version.sdkInt}), ${androidInfo.manufacturer} ${androidInfo.model}";
    } else if (Platform.isIOS) {
      IosDeviceInfo iosInfo = await deviceInfo.iosInfo;
      _osVersion = "${iosInfo.systemName}: ${iosInfo.systemVersion}";
      _platformVersion =
          "${iosInfo.systemName}: ${iosInfo.systemVersion}, ${iosInfo.name} ${iosInfo.model}";
    }
  }

  static BuildContext context() {
    return _context;
  }

  static Future<dynamic> consentRequestView(String consentId) async {
    return await Navigator.of(_context)
        .push(MaterialPageRoute(builder: (context) {
      return Scaffold(
        body: ConsentPage(
          consentId: consentId,
          baseAPI: baseAPI(),
        ),
      );
    }));
  }

  static Future<dynamic> notificationView() async {
    return await Navigator.of(_context)
        .push(MaterialPageRoute(builder: (context) {
      return Scaffold(
        body: NotificationsPage(
          baseAPI: baseAPI(),
        ),
      );
    }));
  }

  static Widget consentRequestWidget(String consentId) {
    return ConsentPage(
      consentId: consentId,
      baseAPI: baseAPI(),
    );
  }

  static Widget notificationWidget() {
    return NotificationsPage(
      baseAPI: baseAPI(),
    );
  }

  static String appPrefix() {
    if (_appPrefix == null) {
      _appPrefix = _config["app-prefix"];
    }
    return _appPrefix!;
  }

  static String pamServer() {
    if (_pamServer == null) {
      _pamServer = _config["pam-server"];
    }
    return _pamServer!;
  }

  static String publicDBAlias() {
    if (_publicDBAlias == null) {
      _publicDBAlias = _config["public-db-alias"];
    }
    return _publicDBAlias!;
  }

  static String loginDBAlias() {
    if (_publicDBAlias == null) {
      _publicDBAlias = _config["public-db-alias"];
    }
    return _publicDBAlias!;
  }

  static String baseAPI() {
    if (_baseAPI == null) {
      _baseAPI = _config["base-api"];
    }
    return _baseAPI!;
  }

  static String trackingConsentId() {
    if (_trackingConsentId == null) {
      _trackingConsentId = _config["tracking-consent-id"];
    }
    return _trackingConsentId!;
  }

  static String contactingConsentId() {
    if (_contactingConsentId == null) {
      _contactingConsentId = _config["contacting-consent-id"];
    }
    return _contactingConsentId!;
  }

  static void listen(
      String eventName, void Function(Map<String, dynamic>) callback) {
    if (eventName.toLowerCase() == PamStandardCallback.on_token) {
      onTokenListener.add(callback);
    } else if (eventName.toLowerCase() ==
        PamStandardCallback.on_message_opened) {
      onMessageOpenedListener.add(callback);
    } else if (eventName.toLowerCase() == PamStandardCallback.on_message) {
      onMessageListener.add(callback);
    }
  }

  static void dispatch(String eventName, Map<String, dynamic> payload) {
    if (eventName.toLowerCase() == PamStandardCallback.on_consent) {
      onConsentListener.forEach((cb) {
        cb.call(payload);
      });
    } else if (eventName.toLowerCase() == PamStandardCallback.on_token) {
      onTokenListener.forEach((cb) {
        cb.call(payload);
      });
    } else if (eventName.toLowerCase() ==
        PamStandardCallback.on_message_opened) {
      onMessageOpenedListener.forEach((cb) {
        cb.call(payload);
      });
    } else if (eventName.toLowerCase() == PamStandardCallback.on_message) {
      onMessageListener.forEach((cb) {
        cb.call(payload);
      });
    }
  }

  static Future<void> appReady({
    required BuildContext context,
    Map<String, dynamic>? initialMessage,
    Widget Function(String id)? productProductDetail,
  }) async {
    if (!_isAppReady) {
      _isAppReady = true;
      _context = context;
      _productDetail = productProductDetail;
      _track(PamStandardEvent.app_launch);
      _pendingMessage.forEach((payload) {
        dispatch(PamStandardCallback.on_message, payload);
      });
      _pendingMessage.clear();
    }

    if (await PamNotification.isPamMessageReceived(initialMessage) &&
        initialMessage != null) {
      PamMessageModel pamMessage =
          PamMessageModel.fromJson(appPrefix(), initialMessage);
      if (productProductDetail != null &&
          pamMessage.PageTitle == "product_detail") {
        await Navigator.of(_context).pushAndRemoveUntil(
            MaterialPageRoute(
                builder: (context) => Scaffold(
                      body: productProductDetail(pamMessage.Id ?? ""),
                    )),
            (route) => false);
      }
    }
  }

  static Future<void> askNotificationPermission() async {
    return PamNotification.askNotificationPermission();
  }

  static dynamic _getFromSharedPref(String key) {
    return _sharedPreferences.get(key);
  }

  static String? _getStringFromSharedPref(String key) {
    return _sharedPreferences.getString(key);
  }

  static int? _getIntFromSharedPref(String key) {
    return _sharedPreferences.getInt(key);
  }

  static Future<bool> _saveStringToSharedPref(String key, String value) {
    return _sharedPreferences.setString(key, value);
  }

  static Future<bool> _saveIntToSharedPref(String key, int value) {
    return _sharedPreferences.setInt(key, value);
  }

  static Future<bool> _removeFromSharedPref(String key) {
    return _sharedPreferences.remove(key);
  }

  static Future<String> _getSessionId() async {
    int? expire = _getIntFromSharedPref("session_expire_in_mils");
    String? session = _getStringFromSharedPref("session_id");

    if ((expire == null || session == null) ||
        (DateTime.now().millisecondsSinceEpoch - expire > (1000 * 60 * 60))) {
      int newExpire = DateTime.now().millisecondsSinceEpoch + 1000 * 60 * 60;
      session = Uuid().v4();

      await _saveIntToSharedPref("session_expire_in_mils", newExpire);
      await _saveStringToSharedPref("session_id", session);
    } else {
      int newExpire = DateTime.now().millisecondsSinceEpoch + 1000 * 60 * 60;

      await _saveIntToSharedPref("session_expire_in_mils", newExpire);
      await _saveStringToSharedPref("session_id", session);
    }

    return session;
  }

  static String? _getContactId() {
    if (_contactId == null) {
      _contactId =
          _getStringFromSharedPref(PamStandardSharePreferenceKey.contact_id);
    }
    return _contactId!;
  }

  static String? _getCustomerId() {
    if (_customerId == null) {
      _customerId =
          _getStringFromSharedPref(PamStandardSharePreferenceKey.customer_id);
    }
    return _customerId!;
  }

  static Widget Function(String id)? productDetail() {
    return _productDetail;
  }

  static Future<void> userLogin(String customerId) async {
    await _saveStringToSharedPref(
        PamStandardSharePreferenceKey.customer_id, customerId);
    _track(
      PamStandardEvent.login,
      payload: {"customer": customerId},
    );
  }

  static Future<void> userLogout() async {
    Map<String, dynamic> payload = {};
    if (Platform.isAndroid) {
      payload["_delete_media"] = {"android_notification": ""};
    } else if (Platform.isIOS) {
      payload["_delete_media"] = {"ios_notification": ""};
    }

    _track(
      PamStandardEvent.logout,
      payload: payload,
    );
    await _removeFromSharedPref(PamStandardSharePreferenceKey.customer_id);
    await _removeFromSharedPref(PamStandardSharePreferenceKey.contact_id);
  }

  static Future<void> trackPageView({
    required String url,
    required String title,
    Map<String, dynamic>? others,
  }) async {
    return _track(PamStandardEvent.page_view, payload: {
      "page_url": url,
      "page_title": title,
      ...(others ?? Map<String, dynamic>.from({})),
    });
  }

  static Future<void> trackFavourite({
    required String id,
    required String title,
    required double price,
    Map<String, dynamic>? others,
  }) async {
    return _track(PamStandardEvent.favourite, payload: {
      "page_title": title,
      "product_id": id,
      "product_title": title,
      "product_price": price,
      ...(others ?? Map<String, dynamic>.from({})),
    });
  }

  static Future<void> trackAddToCart({
    required String id,
    String? title,
    double? price,
    int? quantity,
    Map<String, dynamic>? others,
  }) async {
    return _track(PamStandardEvent.add_to_cart, payload: {
      "page_title": title,
      "product_id": id,
      "product_title": title,
      "product_price": price,
      "quantity": quantity ?? 1,
      ...(others ?? Map<String, dynamic>.from({})),
    });
  }

  static Future<void> trackRemoveFromCart({
    required String id,
    String? title,
    double? price,
    int? quantity,
    Map<String, dynamic>? others,
  }) async {
    return _track(PamStandardEvent.remove_from_cart, payload: {
      "page_title": title,
      "product_id": id,
      "product_title": title,
      "product_price": price,
      "quantity": quantity ?? 1,
      ...(others ?? Map<String, dynamic>.from({})),
    });
  }

  static Future<void> trackPurchaseSuccess({
    required List<String> ids,
    required List<String> titles,
    required List<String> categories,
    required double? totalPrice,
    Map<String, dynamic>? others,
  }) async {
    return _track(PamStandardEvent.purchase_success, payload: {
      "product_id": ids.join(","),
      "product_title": titles.join(","),
      "product_cat": categories.join(","),
      "total_price": totalPrice,
      ...(others ?? Map<String, dynamic>.from({})),
    });
  }

  static Future<void> trackSavePush({
    required String token,
    Map<String, dynamic>? others,
  }) async {
    Map<String, dynamic> pl = {};
    if (Platform.isAndroid) {
      pl["android_notification"] = token;
    } else if (Platform.isIOS) {
      pl["ios_notification"] = token;
    }

    dispatch(PamStandardCallback.on_token, {
      "token": token,
    });

    return _track(PamStandardEvent.save_push, payload: {
      ...pl,
      ...(others ?? Map<String, dynamic>.from({})),
    });
  }

  static Future<void> trackCustomEvent({
    required String eventName,
    Map<String, dynamic>? payload,
  }) async {
    return _track(PamStandardEvent.page_view, payload: {
      ...(payload ?? Map<String, dynamic>.from({})),
    });
  }

  static void _track(
    String eventName, {
    Map<String, dynamic>? payload,
    bool deleteLoginContactAfterPost = false,
  }) {
    return _pamQueueManager.enqueue(eventName,
        payload: payload,
        deleteLoginContactAfterPost: deleteLoginContactAfterPost);
  }

  static bool _isUserLogin() {
    if (_contactId == null) {
      _contactId =
          _getStringFromSharedPref(PamStandardSharePreferenceKey.customer_id);
      return _contactId != null;
    }
    return true;
  }

  static Future<void> _postTracker(
    String eventName,
    Map<String, dynamic> payload, {
    bool deleteLoginContactAfterPost = false,
  }) async {
    String url = "${pamServer()}/trackers/events";
    Map<String, dynamic> body = {
      "event": eventName,
      "platform": _platformVersion,
    };
    Map<String, dynamic> formField = {
      "os_version": _osVersion,
      "app_version": _appVersion,
      "_session_id": await _getSessionId(),
    };

    if (_getStringFromSharedPref(
            PamStandardSharePreferenceKey.login_contact_id) !=
        null) {
      formField["_contact_id"] = _getStringFromSharedPref(
          PamStandardSharePreferenceKey.login_contact_id);
    } else if (_getStringFromSharedPref(
            PamStandardSharePreferenceKey.contact_id) !=
        null) {
      formField["_contact_id"] =
          _getStringFromSharedPref(PamStandardSharePreferenceKey.contact_id);
    }

    payload.forEach((key, value) {
      if (key == "page_url" || key == "page_title") {
        body[key] = value;
      } else {
        formField[key] = value;
      }
    });

    if (_isUserLogin()) {
      formField["_database"] = loginDBAlias();
      formField["customer"] = _getCustomerId();
    } else {
      formField["_database"] = publicDBAlias();
    }

    body["form_fields"] = formField;

    if (_enableLog) {
      print("🦄 : POST Event = 🍀$eventName🍀");
      print("🦄 : Payload");
      print(
          "🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉\n$body\n🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉🍉\n\n");
    }

    Response response = await Requester.post(url, body);
    if (_enableLog) {
      print("Track response is ${response.body}\n");
    }

    if (response.statusCode == 200) {
      IPamResponse pamResponse =
          IPamResponse.fromJson(json.decode(response.body));
      if (pamResponse.ContactId != null) {
        if (_isUserLogin()) {
          _saveStringToSharedPref(
              PamStandardSharePreferenceKey.login_contact_id,
              pamResponse.ContactId!);
        } else {
          _saveStringToSharedPref(
              PamStandardSharePreferenceKey.contact_id, pamResponse.ContactId!);
        }
      }
    } else {
      if (_enableLog) {
        print("Track error ${response.body}\n");
      }
    }

    if (deleteLoginContactAfterPost) {
      _customerId = null;
      _removeFromSharedPref(PamStandardSharePreferenceKey.customer_id);
      _removeFromSharedPref(PamStandardSharePreferenceKey.login_contact_id);
    }
    _pamQueueManager.nextQueue();

    if (_enableLog) {
      print("Track has been send\n");
    }
  }
}
